//
//  PSMSlideshow.m
//  PSMSlideshow
//
//  Created by John Pannell on 9/14/06.
//  Copyright 2006 Positive Spin Media. All rights reserved.
//

#import "PSMSlideshow.h"
#import "PSMSlideshowImage.h"
#import "PSMSlideshowInternalDataSource.h"

@interface PSMSlideshow (Private)
// intenal image management
- (PSMSlideshowImage *)incomingImage;
- (void)setIncomingImage:(PSMSlideshowImage *)image;
- (PSMSlideshowImage *)outgoingImage;
- (void)setOutgoingImage:(PSMSlideshowImage *)image;

// timing
- (void)animate:(NSTimer *)timer;
- (void)setReferenceDate:(NSDate *)date;
- (void)prepareTransition;
- (void)loadImageAtIndex:(NSNumber *)index;
- (void)startTransitionWithImage:(PSMSlideshowImage *)image;
- (int)randomIndex;

@end

@implementation PSMSlideshow

+ (NSArray *)availableTransitionNames
{
    // get contents of framework resource folder
    NSMutableArray *names = [NSMutableArray array];
    NSArray *paths = [[NSBundle bundleForClass:[PSMSlideshow class]] pathsForResourcesOfType:@"qtz" inDirectory:nil];
    NSEnumerator *e = [paths objectEnumerator];
    NSString *path;
    while(path = [e nextObject]){
        [names addObject:[[path stringByDeletingPathExtension] lastPathComponent]];
    }
    return names;
}

+ (NSArray *)directionOptionsForTransitionName:(NSString *)transitionName
{
    // find associated plist file for given transition name
    NSString *infoPath = [[NSBundle bundleForClass:[PSMSlideshow class]] pathForResource:transitionName ofType:@"plist"];
    
    if(!infoPath)
        return nil;
    
    NSDictionary *dict = [NSDictionary dictionaryWithContentsOfFile:infoPath];
    
    return [dict objectForKey:@"directionOptions"];
}

#pragma mark -
#pragma mark Creation/Destruction

- (id)initWithImagePaths:(NSArray *)paths
{
    // create internal data source with image paths
    PSMSlideshowInternalDataSource *ds = [[[PSMSlideshowInternalDataSource alloc] initWithFilePaths:paths] autorelease];
    //PSMSlideshowInternalDataSource *ds = [[[PSMSlideshowInternalDataSource alloc] initWithTestPaths:paths storageFormat:5] autorelease];
    return [self initWithDataSource:ds];
}

- (id)initWithContentsOfFolder:(NSString *)path includeSubfolders:(BOOL)flag
{
    NSFileManager *fm = [NSFileManager defaultManager];
    BOOL isDirectory = NO;
    NSMutableArray *fileNames = [NSMutableArray array];
    
    // create list of file names
    if ([fm fileExistsAtPath:path isDirectory:&isDirectory] && isDirectory){
        NSDirectoryEnumerator *myDirectoryEnumerator = [fm enumeratorAtPath:path];
        NSString *myFile = nil;
        while (myFile = [myDirectoryEnumerator nextObject]) {
            NSString *myCompletePath = [[path mutableCopy] stringByAppendingPathComponent:myFile];
            // subdirectories?
            if(!flag){
                [myDirectoryEnumerator skipDescendents];
            }
            [fileNames addObject:myCompletePath];
        }
    }
    
    return [self initWithImagePaths:fileNames];
}

// designated initializer
- (id)initWithDataSource:(id <PSMSlideshowDataSource>)dataSource
{
    if(self = [super init]){
        [dataSource retain];
        slides = dataSource;
        
        // set default variables for config
        transitionDuration = 3.0;
        displayDuration = 5.0;
        useKenBurnsEffect = NO;
        scaleToFill = NO;
        randomizeImages = NO;
        autoAdvance = NO;
        repeatShow = NO;
        transitionName = [[NSString alloc] initWithString:@"Cube"];
        transitionDirection = -2;
        isPaused = NO;
        
        slideIndex = 0;
        slideshowView = [[QCView alloc] initWithFrame:NSMakeRect(0.0, 0.0, 100.0, 100.0)];
        
        animationTimer = [NSTimer timerWithTimeInterval:(1.0/60.0) target:self selector:@selector(animate:) userInfo:nil repeats:YES];
        [[NSRunLoop currentRunLoop] addTimer:animationTimer forMode:NSDefaultRunLoopMode];
        [[NSRunLoop currentRunLoop] addTimer:animationTimer forMode:NSEventTrackingRunLoopMode];
        [self setReferenceDate:[NSDate date]];
        nextSlideIsReady = YES;
        currentTransitionTime = 0.0;
        
        viewedSlideIndices = [[NSMutableArray alloc] init];
        srandomdev();
    }
    
    return self;
}

#pragma mark -
#pragma mark Accessors

- (float)transitionDuration
{
    return transitionDuration;
}

- (void)setTransitionDuration:(float)value
{
    transitionDuration = value;
}

- (float)displayDuration
{
    return displayDuration;
}

- (void)setDisplayDuration:(float)value
{
    displayDuration = value;
}

- (BOOL)useKenBurnsEffect
{
    return useKenBurnsEffect;
}

- (void)setUseKenBurnsEffect:(BOOL)flag
{
    useKenBurnsEffect = flag;
}

- (BOOL)scaleToFill
{
    return scaleToFill;
}

- (void)setScaleToFill:(BOOL)flag
{
    scaleToFill = flag;
    [slideshowView setValue:[NSNumber numberWithBool:flag] forInputKey:@"scaleToFill"];
}

- (BOOL)randomizeImages
{
    return randomizeImages;
}

- (void)setRandomizeImages:(BOOL)flag
{
    randomizeImages = flag;
}

- (BOOL)autoAdvance
{
    return autoAdvance;
}

- (void)setAutoAdvance:(BOOL)flag
{
    autoAdvance = flag;
}

- (BOOL)repeatShow
{
    return repeatShow;
}

- (void)setRepeatShow:(BOOL)flag
{
    repeatShow = flag;
}

- (NSString *)transitionName
{
    return transitionName;
}

- (void)setTransitionName:(NSString *)name
{
    [name retain];
    [transitionName release];
    transitionName = name;
    
    // load composition
    [slideshowView loadCompositionFromFile:[[[[NSBundle bundleForClass:[PSMSlideshow class]] resourcePath] stringByAppendingPathComponent:name] stringByAppendingPathExtension:@"qtz"]];
    
    // reset the inputs
    [slideshowView setValue:(id)[[self outgoingImage] fastImage] forInputKey:@"imageOut"];
    [slideshowView setValue:[NSNumber numberWithFloat:[outgoingImage positionX]] forInputKey:@"imageOutX"];
    [slideshowView setValue:[NSNumber numberWithFloat:[outgoingImage positionY]] forInputKey:@"imageOutY"];
    [slideshowView setValue:[NSNumber numberWithFloat:[outgoingImage positionZ]] forInputKey:@"imageOutZ"];
    [slideshowView setValue:(id)[[self incomingImage] fastImage] forInputKey:@"imageIn"];
    [slideshowView setValue:[NSNumber numberWithFloat:[incomingImage positionX]] forInputKey:@"imageInX"];
    [slideshowView setValue:[NSNumber numberWithFloat:[incomingImage positionY]] forInputKey:@"imageInY"];
    [slideshowView setValue:[NSNumber numberWithFloat:[incomingImage positionZ]] forInputKey:@"imageInZ"];
}

- (int)transitionDirection
{
    return transitionDirection;
}

- (void)setTransitionDirection:(int)value
{
    transitionDirection = value;
    [slideshowView setValue:[NSNumber numberWithInt:value] forInputKey:@"direction"];
}

- (int)slideCount
{
    return [slides numberOfSlideshowItems];
}

- (int)currentSlideIndex
{
    return slideIndex;
}

- (BOOL)isPaused
{
    return isPaused;
}

- (PSMSlideshowImage *)incomingImage
{
    return incomingImage;
}

- (void)setIncomingImage:(PSMSlideshowImage *)image
{
    [image retain];
    [incomingImage release];
    incomingImage = image;
}

- (PSMSlideshowImage *)outgoingImage
{
    return outgoingImage;
}

- (void)setOutgoingImage:(PSMSlideshowImage *)image
{
    [image retain];
    [outgoingImage release];
    outgoingImage = image;
}

#pragma mark -
#pragma mark Control

- (void)load
{
    [slideshowView loadCompositionFromFile:[[[[NSBundle bundleForClass:[PSMSlideshow class]] resourcePath] stringByAppendingPathComponent:transitionName] stringByAppendingPathExtension:@"qtz"]];
    [slideshowView startRendering];
    
    // get started - transition from nothing to first slide
    slideIndex = 0;
    [self showNextSlide];
    
}

- (void)rewind
{
    slideIndex = 0;
    PSMSlideshowImage *firstImage = [[PSMSlideshowImage alloc] initWithSlideshowItem:[slides slideshowItemAtIndex:slideIndex] forSlideshow:self];
    
    // move in to out
    [self setOutgoingImage:firstImage];
    [slideshowView setValue:nil forInputKey:@"imageOut"];
    [self setIncomingImage:firstImage];
    [slideshowView setValue:(id)[firstImage image] forInputKey:@"imageIn"];
    // reset transition
    currentTransitionTime = 0.0;
    [slideshowView setValue:[NSNumber numberWithFloat:(currentTransitionTime/transitionDuration)] forInputKey:@"transitionFactor"];
    slideIndex++;
}

- (void)pause
{
    isPaused = YES;
}

- (void)resume
{
    isPaused = NO;
}

- (void)showNextSlide
{
    if(slideIndex >= [slides numberOfSlideshowItems])
        return;
    
    if(!nextSlideIsReady){
        // I'm busy, call back later :-)
        [self performSelector:@selector(showNextSlide) withObject:nil afterDelay:0.1];
        return;
    }
    
    // go
    [self prepareTransition];
    [slideshowView setValue:[NSNumber numberWithInt:transitionDirection] forInputKey:@"direction"];
}

- (void)showPreviousSlide
{
    slideIndex -= 2; // 0 is on deck, -1 is current view, -2 is the one before it.
    if(slideIndex < 0)
        return;
    
    if(!nextSlideIsReady){
        // I'm busy, call back later :-)
        [self performSelector:@selector(showPreviousSlide) withObject:nil afterDelay:0.1];
        return;
    }

    // go
    [self prepareTransition];
    [slideshowView setValue:[NSNumber numberWithInt:(-transitionDirection)] forInputKey:@"direction"];
}

#pragma mark -
#pragma mark Internal

- (void)animate:(NSTimer *)timer
{
    
    // perform transition time increment
    if(!isPaused){
        float interval = [[NSDate date] timeIntervalSinceDate:referenceDate];
        currentTransitionTime += interval;
        [slideshowView setValue:[NSNumber numberWithFloat:(currentTransitionTime/transitionDuration)] forInputKey:@"transitionFactor"];
        
        // ken burns?
        if([self useKenBurnsEffect] && [self autoAdvance]){
            [incomingImage animate];
            [outgoingImage animate];
            [slideshowView setValue:[NSNumber numberWithFloat:[incomingImage positionX]] forInputKey:@"imageInX"];
            [slideshowView setValue:[NSNumber numberWithFloat:[incomingImage positionY]] forInputKey:@"imageInY"];
            [slideshowView setValue:[NSNumber numberWithFloat:[incomingImage positionZ]] forInputKey:@"imageInZ"];
            [slideshowView setValue:[NSNumber numberWithFloat:[outgoingImage positionX]] forInputKey:@"imageOutX"];
            [slideshowView setValue:[NSNumber numberWithFloat:[outgoingImage positionY]] forInputKey:@"imageOutY"];
            [slideshowView setValue:[NSNumber numberWithFloat:[outgoingImage positionZ]] forInputKey:@"imageOutZ"];
        }
        
        // perform auto advance if needed
        if((currentTransitionTime > (transitionDuration + displayDuration)) && autoAdvance && nextSlideIsReady)
            [self showNextSlide];
    }
    
    [self setReferenceDate:[NSDate date]];
    
}

- (void)setReferenceDate:(NSDate *)date
{
    [date retain];
    [referenceDate release];
    referenceDate = date;
}

- (void)prepareTransition
{
    nextSlideIsReady = NO;
    int chosenIndex = slideIndex;
    
    if(randomizeImages) 
        chosenIndex = [self randomIndex];
    
    if(slideIndex < [viewedSlideIndices count]) // user went backward - pick appropriate viewed slide
        chosenIndex = [[viewedSlideIndices objectAtIndex:slideIndex] intValue];
    
    [NSThread detachNewThreadSelector:@selector(loadImageAtIndex:) toTarget:self withObject:[NSNumber numberWithInt:chosenIndex]];
    
    // record usage and advance counter
    if(![viewedSlideIndices containsObject:[NSNumber numberWithInt:chosenIndex]])
        [viewedSlideIndices addObject:[NSNumber numberWithInt:chosenIndex]];
    
    slideIndex++;
    
    // done?
    if(slideIndex >= [slides numberOfSlideshowItems]){
        if(repeatShow){
            slideIndex = 0;
            [viewedSlideIndices removeAllObjects];
        }
    }
}

- (void)loadImageAtIndex:(NSNumber *)index
{
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
    
    PSMSlideshowImage *nextImage = [[PSMSlideshowImage alloc] initWithSlideshowItem:[slides slideshowItemAtIndex:[index intValue]] forSlideshow:self];
    
    [self performSelectorOnMainThread:@selector(startTransitionWithImage:) withObject:nextImage waitUntilDone:NO];
    
    [pool release];
    [NSThread exit];
}
    

- (void)startTransitionWithImage:(PSMSlideshowImage *)image
{
    // move in to out
    [self setOutgoingImage:[self incomingImage]];
    [slideshowView setValue:(id)[[self outgoingImage] fastImage] forInputKey:@"imageOut"];
    [slideshowView setValue:[NSNumber numberWithFloat:[outgoingImage positionX]] forInputKey:@"imageOutX"];
    [slideshowView setValue:[NSNumber numberWithFloat:[outgoingImage positionY]] forInputKey:@"imageOutY"];
    [slideshowView setValue:[NSNumber numberWithFloat:[outgoingImage positionZ]] forInputKey:@"imageOutZ"];
    
    // reset transition
    currentTransitionTime = 0.0;
    nextSlideIsReady = YES;
    [slideshowView setValue:[NSNumber numberWithFloat:(currentTransitionTime/transitionDuration)] forInputKey:@"transitionFactor"];
    
    // load new in
    [self setIncomingImage:image];
    [slideshowView setValue:(id)[[self incomingImage] fastImage] forInputKey:@"imageIn"];
    [slideshowView setValue:[NSNumber numberWithFloat:[incomingImage positionX]] forInputKey:@"imageInX"];
    [slideshowView setValue:[NSNumber numberWithFloat:[incomingImage positionY]] forInputKey:@"imageInY"];
    [slideshowView setValue:[NSNumber numberWithFloat:[incomingImage positionZ]] forInputKey:@"imageInZ"];
    [image release];
}

- (int)randomIndex
{
    int i, slideCount = [slides numberOfSlideshowItems];
    NSMutableSet *possibleIndices = [[NSMutableSet alloc] initWithCapacity:slideCount];
    for(i = 0; i < slideCount; i++){
        [possibleIndices addObject:[NSNumber numberWithInt:i]];
    }
    
    NSSet *usedIndices = [NSSet setWithArray:viewedSlideIndices];
    [possibleIndices minusSet:usedIndices];
    
    NSArray *hat = [NSArray arrayWithArray:[possibleIndices allObjects]];
    int rabbit = random() % [hat count];
    
    return [[hat objectAtIndex:rabbit] intValue];
}


#pragma mark -
#pragma mark View

- (QCView *)slideshowView
{
    return slideshowView;
}



@end
