//
//  PSMSlideshowImage.m
//  PSMSlideshow
//
//  Created by John Pannell on 9/13/06.
//  Copyright 2006 Positive Spin Media. All rights reserved.
//

#import <Quartz/Quartz.h>
#import "PSMSlideshowImage.h"

@interface PSMSlideshowImage (Private)

- (id)initWithImageRef:(CGImageRef)ref;
- (id)initWithImage:(NSImage *)image;

- (CGImageRef)imageFromPath:(NSString *)path;
- (CGImageRef)imageFromURL:(NSURL *)url;
- (CGImageRef)imageFromFileWrapper:(NSFileWrapper *)wrapper;
- (NSImage *)imageFromPDFPage:(PDFPage *)page;
- (CVPixelBufferRef)fastImageFromCGImage:(CGImageRef)image;
- (CVPixelBufferRef)fastImageFromNSImage:(NSImage *)image;

CGImageRef PSMCreateReasonablySizedImage(CGImageSourceRef source, float viewWidth, float viewHeight);
@end

@implementation PSMSlideshowImage

#pragma mark -
#pragma mark Creation/Destruction

- (id)init
{
    if(self = [super init]){
        if([slideshow useKenBurnsEffect] && [slideshow autoAdvance]){
            // generate random positions and vectors for animation
            srandomdev();
            currentX = ((random() % 200) - 100) / 1000.0; // -.100 thru +.100
            currentY = ((random() % 200) - 100) / 1000.0; // -.100 thru +.100
            currentZ = ((random() % 100) + 500) / 1000.0; // .500 thru +.700
            float numberOfFrames = 60 * ([slideshow displayDuration] + [slideshow transitionDuration]);
            while(sqrtf(animationVectorX * animationVectorX * animationVectorY *animationVectorY) < .00000001){
                animationVectorX = ((random() % 300) - 150) / 1000.0 / numberOfFrames; // -.150 thru +.150
                animationVectorY = ((random() % 300) - 150) / 1000.0 / numberOfFrames; // -.150 thru +.150
            }
            animationVectorZ = ((random() % 200) - 50) / 1000.0 / numberOfFrames; // -.05 thru +.150
        } else {
            currentX = 0;
            currentY = 0;
            currentZ = 0;
            animationVectorX = 0;
            animationVectorY = 0;
            animationVectorZ = 0;
        }
    }
    
    return self;
}
    

- (id)initWithImageRef:(CGImageRef)ref
{
    sourceImage = (id)ref;
    fastImage = [self fastImageFromCGImage:(CGImageRef)sourceImage];
    return [self init];
}

- (id)initWithImage:(NSImage *)image
{
    sourceImage = (id)image;
    fastImage = [self fastImageFromNSImage:image];
    return [self init];
}

// designated initializer
- (id)initWithSlideshowItem:(id)item forSlideshow:(PSMSlideshow *)sender
{
    slideshow = sender;
    
    // item can be NSImage, NSString (to specify a path name), NSURL, NSFileWrapper, CGImage, or PDFPage.
    if([[item class] isSubclassOfClass:[NSString class]]){
        return [self initWithImageRef:[self imageFromPath:item]];
    } else if([[item class] isSubclassOfClass:[NSURL class]]){
        return [self initWithImageRef:[self imageFromURL:item]];
    } else if([[item class] isSubclassOfClass:[NSImage class]]){
        return [self initWithImage:item];
    } else if([[item class] isSubclassOfClass:[NSFileWrapper class]]){
        return [self initWithImageRef:[self imageFromFileWrapper:item]];
    } else if(CFGetTypeID(item) == CGImageGetTypeID()){ // CGImage
        return [self initWithImageRef:(CGImageRef)item];
    } else if([[item class] isSubclassOfClass:[PDFPage class]]){
        return [self initWithImage:[self imageFromPDFPage:item]];
    }
    return nil;
}

- (void)dealloc
{
    [name release];
    if([[sourceImage class] isSubclassOfClass:[NSImage class]]){
        [sourceImage release];
    } else {
        CFRelease(sourceImage);
    }
    CVPixelBufferRelease(fastImage);
    [super dealloc];
}
        


#pragma mark -
#pragma mark Accessors

- (id)image
{
    return sourceImage;
}

- (CVPixelBufferRef)fastImage
{
    return fastImage;
}

- (float)positionX
{
    return currentX;
}

- (float)positionY
{
    return currentY;
}

- (float)positionZ
{
    return currentZ;
}

#pragma mark -
#pragma mark Functionality

- (void)animate
{
    currentX += animationVectorX;
    currentY += animationVectorY;
    currentZ += animationVectorZ;
}

#pragma mark -
#pragma mark Utility

- (CGImageRef)imageFromPath:(NSString *)path
{
    CGImageRef image = NULL;
    
    CFURLRef url = CFURLCreateFromFileSystemRepresentation(kCFAllocatorDefault,(void *)[path fileSystemRepresentation],[path length],FALSE);
    
    if(!url){
        NSLog(@"Could not create URL from path");
        return NULL;
    }
    
    image = [self imageFromURL:(NSURL *)url];
    CFRelease(url);
    
    return image;
}

- (CGImageRef)imageFromURL:(NSURL *)url
{
    CGImageRef image = NULL;
    CGImageSourceRef imageSource = NULL;
    
    imageSource = CGImageSourceCreateWithURL((CFURLRef)url,NULL);
    if(!imageSource){
        NSLog(@"Could not create image source from URL");
        return NULL;
    }
    
    float viewWidth = [[slideshow slideshowView] frame].size.width;
    float viewHeight = [[slideshow slideshowView] frame].size.height;
    
    image = PSMCreateReasonablySizedImage(imageSource, viewWidth, viewHeight);
    
    CFRelease(imageSource);
    
    return image;
}

- (CGImageRef)imageFromFileWrapper:(NSFileWrapper *)wrapper
{
    CGImageRef image = NULL;
    CGImageSourceRef imageSource = NULL;
    
    imageSource = CGImageSourceCreateWithData((CFDataRef)[wrapper regularFileContents], NULL);
    if(!imageSource){
        NSLog(@"Could not create image source from URL");
        return NULL;
    }
    
    float viewWidth = [[slideshow slideshowView] frame].size.width;
    float viewHeight = [[slideshow slideshowView] frame].size.height;
    
    image = PSMCreateReasonablySizedImage(imageSource, viewWidth, viewHeight);
    
    CFRelease(imageSource);
    
    return image;
}

- (NSImage *)imageFromPDFPage:(PDFPage *)page
{
    NSRect rect = [page boundsForBox:kPDFDisplayBoxMediaBox];
    NSImage *image = [[NSImage alloc] initWithSize:rect.size];
    [image setScalesWhenResized:YES];
    [image lockFocus];
    [[NSColor whiteColor] set];
    NSRectFill(rect);
    [page drawWithBox:kPDFDisplayBoxMediaBox];
    [image unlockFocus];
    
    float viewWidth = [[slideshow slideshowView] frame].size.width;
    float viewHeight = [[slideshow slideshowView] frame].size.height;
    float imageWidth = [image size].width;
    float imageHeight = [image size].height;
    
    if((imageWidth > viewWidth) || (imageHeight > viewHeight)){
        NSSize imageSize = [image size];
        float scaleFactor = (imageWidth/viewWidth > imageHeight/viewHeight) ? (viewWidth/imageWidth) : (viewHeight/imageHeight);
        NSAffineTransform *t = [NSAffineTransform transform];
        [t scaleBy:scaleFactor];
        [image setSize:[t transformSize:imageSize]];
    }
    
    return image;
}

- (CVPixelBufferRef)fastImageFromCGImage:(CGImageRef)image
{
    CVPixelBufferRef buffer = NULL;
    
    // config
    size_t width = CGImageGetWidth(image);
    size_t height = CGImageGetHeight(image);
    size_t bitsPerComponent = 8; // CGImageGetBitsPerComponent(image);
    CGColorSpaceRef cs = CGColorSpaceCreateWithName(kCGColorSpaceGenericRGB);
    CGBitmapInfo bi = kCGImageAlphaNoneSkipFirst; // CGImageGetBitmapInfo(image);
    NSDictionary *d = [NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithBool:YES], kCVPixelBufferCGImageCompatibilityKey, [NSNumber numberWithBool:YES], kCVPixelBufferCGBitmapContextCompatibilityKey, nil];
    CGRect imageRect = CGRectMake(0.0, 0.0, width, height);
        
    // create pixel buffer
    CVPixelBufferCreate(kCFAllocatorDefault, width, height, k32ARGBPixelFormat, (CFDictionaryRef)d, &buffer);
    CVPixelBufferLockBaseAddress(buffer, 0);
    void *rasterData = CVPixelBufferGetBaseAddress(buffer);
    size_t bytesPerRow = CVPixelBufferGetBytesPerRow(buffer);
    
    // context to draw in, set to pixel buffer's address
    CGContextRef ctxt = CGBitmapContextCreate(rasterData, width, height, bitsPerComponent, bytesPerRow, cs, bi);
    if(ctxt == NULL){
        NSLog(@"could not create context");
        return NULL;
    }

    // draw
    CGContextDrawImage(ctxt, imageRect, image);
    CVPixelBufferUnlockBaseAddress(buffer, 0);
    
    CFRelease(ctxt);
    
    return buffer;
}

- (CVPixelBufferRef)fastImageFromNSImage:(NSImage *)image
{
    CVPixelBufferRef buffer = NULL;
    
    // config
    size_t width = [image size].width;
    size_t height = [image size].height;
    size_t bitsPerComponent = 8; // CGImageGetBitsPerComponent(image);
    CGColorSpaceRef cs = CGColorSpaceCreateWithName(kCGColorSpaceGenericRGB);
    CGBitmapInfo bi = kCGImageAlphaNoneSkipFirst; // CGImageGetBitmapInfo(image);
    NSDictionary *d = [NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithBool:YES], kCVPixelBufferCGImageCompatibilityKey, [NSNumber numberWithBool:YES], kCVPixelBufferCGBitmapContextCompatibilityKey, nil];
    
    // create pixel buffer
    CVPixelBufferCreate(kCFAllocatorDefault, width, height, k32ARGBPixelFormat, (CFDictionaryRef)d, &buffer);
    CVPixelBufferLockBaseAddress(buffer, 0);
    void *rasterData = CVPixelBufferGetBaseAddress(buffer);
    size_t bytesPerRow = CVPixelBufferGetBytesPerRow(buffer);
    
    // context to draw in, set to pixel buffer's address
    CGContextRef ctxt = CGBitmapContextCreate(rasterData, width, height, bitsPerComponent, bytesPerRow, cs, bi);
    if(ctxt == NULL){
        NSLog(@"could not create context");
        return NULL;
    }
    
    // draw
    NSGraphicsContext *nsctxt = [NSGraphicsContext graphicsContextWithGraphicsPort:ctxt flipped:NO];
    [NSGraphicsContext saveGraphicsState];
    [NSGraphicsContext setCurrentContext:nsctxt];
    [image compositeToPoint:NSMakePoint(0.0, 0.0) operation:NSCompositeCopy];
    [NSGraphicsContext restoreGraphicsState];
    
    CVPixelBufferUnlockBaseAddress(buffer, 0);
    
    CFRelease(ctxt);
    
    return buffer;
}

CGImageRef PSMCreateReasonablySizedImage(CGImageSourceRef source, float viewWidth, float viewHeight)
{
    CGImageRef image = NULL;
    image = CGImageSourceCreateImageAtIndex(source, 0, NULL);
    
    // resize if necessary
    size_t imageWidth = CGImageGetWidth(image);
    size_t imageHeight = CGImageGetHeight(image);
    
    if((imageHeight > viewHeight) || (imageWidth > viewWidth)){
        CFRelease(image);
        int maxSize = floor(viewWidth > viewHeight ? viewWidth : viewHeight);
        CFNumberRef maxSizeRef = CFNumberCreate(kCFAllocatorDefault, kCFNumberIntType, &maxSize);
        CFStringRef keys[2];
        CFTypeRef values[2];
        CFDictionaryRef options = NULL;
        
        keys[0] = kCGImageSourceCreateThumbnailFromImageAlways;
        values[0] = (CFTypeRef)kCFBooleanTrue;
        keys[1] = kCGImageSourceThumbnailMaxPixelSize;
        values[1] = (CFTypeRef)maxSizeRef;
        options = CFDictionaryCreate(kCFAllocatorDefault, (const void **)keys, (const void **)values, 2, &kCFTypeDictionaryKeyCallBacks, &kCFTypeDictionaryValueCallBacks);
        
        image = CGImageSourceCreateThumbnailAtIndex(source, 0, options);
    }
    
    return image;
}

@end
